#ifndef _LWRPC_INCLUDED
#define _LWRPC_INCLUDED

#include <pthread.h>
#include <mpi.h>

#define MODE_MS		0
#define MODE_SPMD	1

void lwrpc_package_init2 (void (*entry)(int, char **), int argc, char *argv[], int mode);
void lwrpc_package_init (int argc, char *argv[], int mode);
void lwrpc_package_reinit();
void lwrpc_package_suspend();
void lwrpc_reset_statistics();

typedef double lwrpc_time_t;
lwrpc_time_t lwrpc_gettime();

int lwrpc_i_worker_id(void);
int lwrpc_i_num_workers();
int lwrpc_worker_id();
int lwrpc_num_workers();

void lwrpc_yield(void);

#define CALL_BY_VAL				(int)(0x0000)
#define CALL_BY_PTR				(int)(0x0001)
#define CALL_BY_REF				(int)(0x0002)
#define CALL_BY_RES				(int)(0x0003)
#define CALL_BY_RMA				(int)(0x0004)
#define CALL_BY_VAD				(int)(0x0005)
#define CALL_BY_COP				(int)(0x0006)

#define OP_NULL					(int)(0x0000)
#define OP_SUM					(int)(0x0100)
#define OP_PROD					(int)(0x0200)

void lwrpc_init();
void lwrpc_waitall();
void lwrpc_create (int queue, void (*f) (), int narg, ...);
void lwrpc_create_local (void (*f) (), int narg, ...);

void lwrpc_spmd (int mi, void (*f) (), int narg, ...);
void lwrpc_spmd_barrier();

void lwrpc_op_red(void *global, void *local, int n, MPI_Datatype dtype, int redop);

int lwrpc_node_id();
int lwrpc_num_nodes();

void lwrpc_broadcast(void *a, long count, MPI_Datatype dtype);

void thread_sleep(int ms);

int lwrpc_sched_nextcpu(int cpu, int stride);


#endif

