/* This file is part of the LWRPC library */
#ifndef _LWRPC_MPI_INTERNAL_INCLUDED
#define _LWRPC_MPI_INTERNAL_INCLUDED

#include <mpi.h>

void enter_comm_cs();
void leave_comm_cs();

void _lwrpc_comm_pre_init();
void _lwrpc_comm_init();

int global_thread_id_to_node_id(int global_thread_id);
int local_thread_id_to_global_thread_id(int local_thread_id);
int global_thread_id_to_local_thread_id(int global_thread_id);

int _lwrpc_total_num_threads();

void *server_loop (void *arg);
void terminate_workers();
void notify_spmd_barrier();
void shutdown_server_thread();
void start_server_thread();
int is_finalized();

void send_descriptor(int, lwrpc_t *, int);
void direct_send_descriptor(int dummy, int sourcenode, int sourcevpid, lwrpc_t *desc);
void receive_arguments(lwrpc_t *work, int tag);
int receive_descriptor(int node, lwrpc_t *work);


#define MAX_NODES					128

#define TERMINATE_LOCAL_SERVER_THREAD			(-20)
#define TERMINATE_WORKER_THREADS			(-21)
#define ASYNCHRONOUS_STEALING_REQUEST			(-22)
#define DIRECT_SYNCHRONOUS_STEALING_REQUEST		(-23)
#define DISABLE_INTERNODE_STEALING      		(-24)
#define ENABLE_INTERNODE_STEALING			(-25)
#define RESET_STATISTICS				(-26)
#define SPMD_BARRIER					(-27)


#define LWRPC_NORMAL                              	(-39)
#define LWRPC_ANSWER                              	(-40)
#define LWRPC_NORMAL_ENQUEUE                      	(-41)
#define LWRPC_NO_WORK                             	(-42)

struct node_info {
	int nprocessors;
	int nworkers;
};

extern struct node_info *node_info;

#define Error(msg)						\
{								\
	printf("ERROR in %s: %s\n", __func__, msg);		\
	MPI_Abort(MPI_COMM_WORLD, 1);				\
}

#define Error1(msg,arg)						\
{								\
	printf("ERROR in %s:", __func__);			\
	printf(msg, arg);					\
	printf("\n");						\
	MPI_Abort(MPI_COMM_WORLD, 1);				\
}

#include "mpi_backport.h"

#endif
