/* This file is part of the LWRPC library */
#include <lwrpc_internal.h>
#include <lwrpc.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

static int endianess ()
{
	static int _endianess = -1;

	if (_endianess == -1) {
	        long m = sizeof(m);

	        union {
        	        long n ;
        	        char a[sizeof(m)];
        	} foo;

	        foo.n = 1;
	        _endianess = (foo.a[(m - 1)]);
	}
	return _endianess;
}

extern void _lwrpc_set_work_routine(lwrpc_t *rpc, void (*work)());
extern MPI_Comm comm_out;
void lwrpc_init()
{
	lwrpc_t *self = _lwrpc_self();
	_lwrpc_depadd(self, 1);
}

void lwrpc_waitall()
{
	_lwrpc_block();
}


#define _initialize(rpc)					\
{								\
	rpc->homenode = rpc->sourcenode = lwrpc_node_id();	\
	_lock_init(&rpc->lock);					\
        rpc->target_queue = rpc->vp_id = -1;			\
}

void lwrpc_create(int queue, void (*work)(), int narg, ...)
{
	va_list ap;
	int i;
	lwrpc_t * rpc;
	lwrpc_t *self = _lwrpc_self();

	_lock_acquire(&self->lock);
	if (self->ndep == 0) self->ndep = 1;
	_lock_release(&self->lock);

	_lwrpc_depadd(self, 1);

#ifdef LWRPC_STATS
	if (self->level == 0)
		created[self->vp_id]++;
	else
		created2[self->vp_id]++;
#endif

	rpc = _lwrpc_get_reused_desc();

	_initialize(rpc);

	_lwrpc_set_work_routine(rpc, work);
    	rpc->narg = narg;
	
	rpc->rpc_type = 1;	
	rpc->inter_node = 1;
	rpc->c_routine = 1;
	rpc->parent = self;
	rpc->level = self->level + 1;
	
	if (narg>MAX_LWRPC_ARGS) Error("narg > MAX_LWRPC_ARGS");

	va_start(ap, narg);
	for (i=0; i<narg; i++) {
		rpc->quantity[i] = va_arg (ap, int);
		rpc->dtype[i] = va_arg (ap, MPI_Datatype);
		rpc->callway[i] = va_arg (ap, int);
		rpc->operation[i] = rpc->callway[i] & 0xFF00;
		rpc->callway[i] = rpc->callway[i] & 0x00FF;
		if ((rpc->callway[i] == CALL_BY_VAL)&&(rpc->quantity[i] > 1))
			rpc->callway[i] = CALL_BY_PTR;

	}
	
	for (i=0; i<narg; i++) {
		if (rpc->callway[i] == CALL_BY_VAL) {
			int typesize;
			MPI_Type_size(rpc->dtype[i], &typesize);
			switch (typesize) {
			case 4:
				rpc->localarg[i] = va_arg (ap, INT32);
				if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
				break;
			case 8:
				if (rpc->dtype[i] == MPI_DOUBLE) {
					double d = va_arg(ap, double);
					memcpy(&rpc->localarg[i], &d, 8);
					printf("...");
				}
				else
				rpc->localarg[i] = va_arg (ap, INT64);
				break;
			default:
				Error("typesize not 4 or 8!");
				break;
			}
		}
		else if (rpc->callway[i] == CALL_BY_COP) {
			int typesize;
			MPI_Type_size(rpc->dtype[i], &typesize);
			switch (typesize) {
			case 4:
				rpc->localarg[i] = *va_arg (ap, INT32 *);
				if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
				break;
			case 8:
				rpc->localarg[i] = *va_arg (ap, INT64 *);
				break;
			default:
				Error("typesize not 4 or 8!");
				break;
			}
		}
		else {
			VIRT_ADDR addr = va_arg (ap, VIRT_ADDR);
			rpc->localarg[i] = addr;
			if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
		}
	}

	if (queue == -1)
		lwrpc_to_rq_end(rpc);
	else
		lwrpc_to_lrq_end(queue, rpc);
}

void lwrpc_spmd (int master_included, void (* work) (), int narg, ...)
{
	va_list ap;
	int i;
	lwrpc_t *rpc, *lrpc;
	lwrpc_t *self = _lwrpc_self();
	int nodes = lwrpc_num_nodes();
	int me = lwrpc_node_id();

#ifdef LWRPC_STATS
#endif
	rpc = _lwrpc_get_reused_desc();

	_initialize(rpc);
	_lwrpc_set_work_routine(rpc, work);
	rpc->narg = narg;

	rpc->rpc_type = 3;	
	rpc->c_routine = 1;
	rpc->inter_node = 1;
        rpc->parent = self;

	if (narg>MAX_LWRPC_ARGS) Error("narg > MAX_LWRPC_ARGS");

	va_start(ap, narg);
	for (i=0; i<narg; i++) {
		rpc->quantity[i] = va_arg (ap, int);
		rpc->dtype[i] = va_arg (ap, MPI_Datatype);
		rpc->callway[i] = va_arg (ap, int);
		rpc->operation[i] = rpc->callway[i] & 0xFF00;
		rpc->callway[i] = rpc->callway[i] & 0x00FF;
		if ((rpc->callway[i] == CALL_BY_VAL)&&(rpc->quantity[i] > 1))
			rpc->callway[i] = CALL_BY_PTR;
	}

	for (i=0; i<narg; i++) {
		if (rpc->callway[i] == CALL_BY_VAL) {
			int typesize;
			MPI_Type_size(rpc->dtype[i], &typesize);
			switch (typesize) {
			case 4:
				rpc->localarg[i] = va_arg (ap, INT32);
				if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
				break;
			case 8:
				if (rpc->dtype[i] == MPI_DOUBLE) {
					double d = va_arg(ap, double);
					memcpy(&rpc->localarg[i], &d, 8);
					printf("...");
				}
				else
					rpc->localarg[i] = va_arg (ap, INT64);
				break;
			default:
				Error("typesize not 4 or 8!");
				break;
			}
		}
		else if (rpc->callway[i] == CALL_BY_COP) {
			int typesize;
			MPI_Type_size(rpc->dtype[i], &typesize);
			switch (typesize) {
			case 4:
				rpc->localarg[i] = *va_arg (ap, INT32 *);
				if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
				break;
			case 8:
				rpc->localarg[i] = *va_arg (ap, INT64 *);
				break;
			default:
				Error("typesize not 4 or 8!");
				break;
			}
		}
		else {
			rpc->localarg[i] = va_arg (ap, VIRT_ADDR);	
			if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
		}
	}

	_lwrpc_depadd(self, nodes);
	for (i = 0; i < nodes; i++) {
		if (i == me) continue;

		lrpc = _lwrpc_get_reused_desc();
		memcpy(lrpc, rpc, sizeof(lwrpc_t));
		lwrpc_to_npq_end(i, lrpc);
	}

	if (master_included) {
		_lwrpc_depadd(self, 1);
		lwrpc_to_i_pq_end(rpc);
	}
	else {
		_lwrpc_put_reused_desc(rpc);
	}
	_lwrpc_block();
}


void lwrpc_yield (void)
{
	_lwrpc_yield();
}


void lwrpc_spmd_barrier()
{
	if (execution_mode == MODE_MS) {	
		return;
	}

	barrier_arrived++;   
	if (lwrpc_num_nodes() > 1) {    
		notify_spmd_barrier();
	}
	while (barrier_arrived < tobe_finished) { sched_yield(); lwrpc_yield(); }
	MPI_Barrier(comm_out);
	barrier_arrived = 0;
}

double lwrpc_gettime()
{
	struct timeval t;
	gettimeofday(&t, NULL);
	return (double)t.tv_sec + (double)t.tv_usec*1.0E-6;
}

int lwrpc_node_id()
{
	return mpi_rank;
}

int lwrpc_num_nodes()
{
	return mpi_nodes;
}

int lwrpc_i_num_workers()
{
	return kthreads;
}

int lwrpc_i_worker_id(void)
{
	return _lwrpc_self()->vp_id;
}

int lwrpc_num_workers()
{
	if (lwrpc_num_nodes() > 1)
		return _lwrpc_total_num_threads();
	else
		return lwrpc_i_num_workers();
}

int lwrpc_worker_id(void)
{
	if (lwrpc_num_nodes() > 1)
		return local_thread_id_to_global_thread_id(_lwrpc_self()->vp_id);
	else
		return lwrpc_i_worker_id();
}


struct lwrpc_data *lwrpc_data;


void lwrpc_package_init2(void (*routine)(int, char **), int argc, char *argv[], int mode)
{
	static int lwrpc_initialized = 0;

	if (lwrpc_initialized) {
		if (package_suspended) {
			lwrpc_package_reinit();
		}
		return;
	}
	lwrpc_initialized = 1;

	lwrpc_data = calloc(1, sizeof(struct lwrpc_data));
	execution_mode = mode;
	entry_routine = routine;

	_lwrpc_opt(argc, argv);

	tobe_finished = (mode == MODE_MS)? 1: lwrpc_num_nodes();
	_lwrpc_env_init();
	if (entry_routine != NULL) {
		if ((execution_mode == MODE_SPMD)||(lwrpc_node_id() == 0)) {
			;
		}
	}
	_lwrpc_worker(0);	
}


void lwrpc_package_init(int argc, char *argv[], int mode)
{
	lwrpc_package_init2(NULL, argc, argv, mode);
}

void lwrpc_package_reinit()
{
	start_server_thread();
	block_idle_threads = old_block_idle_threads;
	_thread_signal();
	package_suspended = 0;
}

void lwrpc_package_suspend()
{
	old_block_idle_threads = block_idle_threads;

	/*if (block_idle_threads == 0) block_idle_threads = 1;*/
	if (!non_blocking) 
		if (block_idle_threads == 0) block_idle_threads = 1;
	shutdown_server_thread();
	package_suspended = 1;
	if (execution_mode == MODE_SPMD) {
		enter_comm_cs();
		MPI_Barrier(comm_out);
		leave_comm_cs();
	}
}


void F77_FUNC_(lwrpc_init, LWRPC_INIT)()
{
	lwrpc_init();
}

void F77_FUNC_(lwrpc_waitall, LWRPC_WAITALL)()
{
	lwrpc_waitall();
}

void F77_FUNC_(lwrpc_create, LWRPC_CREATE) (int *queue, void (* work) (), int *narg, ...)
{
	va_list ap;
	int i;
	lwrpc_t * rpc;
	lwrpc_t * self = _lwrpc_self();

	_lock_acquire(&self->lock);
	if (self->ndep == 0) self->ndep = 1;
	_lock_release(&self->lock);

	_lwrpc_depadd(self, 1);

#ifdef LWRPC_STATS
	if (self->level == 0)
		created[self->vp_id]++;
	else
		created2[self->vp_id]++;   
#endif

	rpc = _lwrpc_get_reused_desc();

	_initialize(rpc);
	_lwrpc_set_work_routine(rpc, work);
	rpc->narg = *narg;

	rpc->rpc_type = 1;	
	rpc->inter_node = 1;
	rpc->parent = self;
	rpc->level = self->level + 1;
    
	if (*narg>MAX_LWRPC_ARGS) Error("narg > MAX_LWRPC_ARGS\n");

	va_start(ap, narg);
	for (i=0; i<*narg; i++) {
		rpc->quantity[i] = *va_arg (ap, int *);
		{
			MPI_Fint dt = *va_arg (ap, MPI_Fint *);
			rpc->dtype[i] = MPI_Type_f2c(dt);

		}
		rpc->callway[i] = *va_arg (ap, int *);

		rpc->operation[i] = rpc->callway[i] & 0xFF00;
		rpc->callway[i] = rpc->callway[i] & 0x00FF;
	}

    	for (i=0; i<*narg; i++) {

		if ((rpc->callway[i] == CALL_BY_VAL)&&(rpc->quantity[i]>1)) {	
			printf("WARNING! %s: (arg %d) CALL_BY_VAL is currently used only for single-element arguments, changed to CALL_BY_PTR\n", __func__, i);
			rpc->callway[i] = CALL_BY_PTR;
		}
		if ((rpc->callway[i] == CALL_BY_VAD)&&(rpc->quantity[i]>1)) {	
			printf("WARNING! %s: (arg %d) CALL_BY_VAD is currently used only for single-element arguments, changed to CALL_BY_PTR\n", __func__, i);
			rpc->callway[i] = CALL_BY_PTR;
		}


		if (rpc->callway[i] == CALL_BY_VAL) {	
			int typesize;
			MPI_Type_size(rpc->dtype[i], &typesize);
			switch (typesize) {
			case 4:
				rpc->localarg[i] = *va_arg(ap, INT32 *);    
				if (endianess()) rpc->localarg[i] = (rpc->localarg[i] << 32);
				break;
			case 8:
				rpc->localarg[i] = *va_arg(ap, INT64 *);    
				break;
			default:
				Error("typesize not 4 or 8!");
				break;
			}
		}
		else {
			VIRT_ADDR va;
			va = va_arg (ap, VIRT_ADDR);
			rpc->localarg[i] = va;
		}
	}

	if (*queue == -1)
		lwrpc_to_rq_end(rpc);
	else
		lwrpc_to_lrq_end(*queue, rpc);
}

void F77_FUNC_(lwrpc_spmd, LWRPC_SPMD) (int *master_included, void (* work) (), int *narg, ...)
{
	va_list ap;
	int i;
	int args;
	lwrpc_t *rpc, *lrpc;
	lwrpc_t *self = _lwrpc_self();
	int nodes = lwrpc_num_nodes();
	int me = lwrpc_node_id();
	int _master_included = 0;

	if (master_included == NULL)
		 _master_included = 0;
	else
		 _master_included = *master_included;

	if (narg == NULL)
		args = 0;
	else
		args = *narg;

#ifdef LWRPC_STATS
#endif
	rpc = _lwrpc_get_reused_desc();

	_initialize(rpc);
	_lwrpc_set_work_routine(rpc, work);
	rpc->narg = args;

	rpc->rpc_type = 3;	
	rpc->inter_node = 1;
	rpc->parent = self;
    
	if (args>MAX_LWRPC_ARGS) Error("narg > MAX_LWRPC_ARGS");

	if (args > 0)	va_start(ap, narg);

	for (i=0; i<args; i++) {
		rpc->quantity[i] = *va_arg (ap, int *);
		{
			MPI_Fint dt = *va_arg (ap, MPI_Fint *);
			rpc->dtype[i] = MPI_Type_f2c(dt);

		}
		rpc->callway[i] = *va_arg (ap, int *);
		rpc->operation[i] = rpc->callway[i] & 0xFF00;
		rpc->callway[i] = rpc->callway[i] & 0x00FF;
	}

    	for (i=0; i<args; i++) {
		VIRT_ADDR va;
		va = va_arg (ap, VIRT_ADDR);
		rpc->localarg[i] = va;
	}

	_lwrpc_depadd(self, nodes);
	for (i = 0; i < nodes; i++) {
		if (i == me) continue;

		lrpc = _lwrpc_get_reused_desc();
		memcpy(lrpc, rpc, sizeof(lwrpc_t));
		lwrpc_to_npq_end(i, lrpc);
	}

	if (master_included) {
		_lwrpc_depadd(self, 1);
		lwrpc_to_i_pq_end(rpc);
	}
	else {
		_lwrpc_put_reused_desc(rpc);
	}
	_lwrpc_block();
}

void F77_FUNC_(lwrpc_yield, LWRPC_YIELD) (void)
{
	lwrpc_yield();
}

double F77_FUNC_(lwrpc_gettime, LWRPC_GETTIME) (void)
{
	return lwrpc_gettime();
}

void F77_FUNC_(lwrpc_spmd_barrier, LWRPC_SPMD_BARRIER) (void)
{
	lwrpc_spmd_barrier();
}

int F77_FUNC_(lwrpc_node_id, LWRPC_NODE_ID) (void)
{
	return lwrpc_node_id();
}

int F77_FUNC_(lwrpc_num_nodes, LWRPC_NUM_NODES) (void)
{
	return lwrpc_num_nodes();
}

int F77_FUNC_(lwrpc_i_num_workers, LWRPC_I_NUM_WORKERS) (void)
{
	return lwrpc_i_num_workers();
}

int F77_FUNC_(lwrpc_i_worker_id, LWRPC_I_WORKER_ID) (void)
{
	return lwrpc_i_worker_id();
}

int F77_FUNC_(lwrpc_num_workers, LWRPC_NUM_WORKERS) (void)
{
	return lwrpc_num_workers();
}

int F77_FUNC_(lwrpc_worker_id, LWRPC_WORKER_ID) (void)
{
	return lwrpc_worker_id();
}

void F77_FUNC_(lwrpc_package_init, LWRPC_PACKAGE_INIT) (int *mode)
{
	lwrpc_package_init2(NULL, 0, NULL, *mode);
}

void F77_FUNC_(lwrpc_package_init2, LWRPC_PACKAGE_INIT2) (func_t func, int *mode)
{
	lwrpc_package_init2(func, 0, NULL, *mode);
}

void F77_FUNC_(lwrpc_package_reinit, LWRPC_PACKAGE_REINIT) ()
{
	lwrpc_package_reinit();
}

void F77_FUNC_(lwrpc_package_suspend, LWRPC_PACKAGE_SUSPEND) ()
{
	lwrpc_package_suspend();
}
